#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-23 23:59:53
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-24 01:11:59
"""

from typing import Any


class Range:
    def __init__(self, start: Any, end: Any = None):
        self.start = start
        self.end = end

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Range):
            return False
        if self.end is None and value.end is not None:
            return
        if self.start is None and value.start is not None:
            return False
        return self.start == value.start and self.end == value.end

    def __repr__(self) -> str:
        return f"Range({self.start}, {self.end})"


def rangeToString(range: Range) -> str:
    if range is None:
        return ""
    assert isinstance(range, Range), f"Expected Range, got {type(range)}"
    return (
        f"[{range.start.toString()}:{range.end.toString()}]"
        if range.end is not None
        else f"[{range.start}]"
    )


class DFGNode:
    # Expressions are hard to handle because of the recursive nature

    def __init__(self, name: str = None) -> None:
        self.variable_name: str = str(name)  # could be a Tree object
        self.operation: OPType = None
        self.range: Range = None
        self.children: list = []

    def __repr__(self) -> str:
        return f"Node({self.variable_name})"

    def isVariable(self) -> bool:
        return self.operation == OPType.VARIABLE

    def isConstant(self) -> bool:
        return self.operation == OPType.CONSTANT

    def setRange(self, range: Range):
        self.range = range

    def toString(self) -> str:
        print(f"toString not implemented for {self}, op = {self.operation}")
        raise NotImplementedError

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, DFGNode):
            return False
        if self.operation != value.operation:
            return False
        if self.variable_name != value.variable_name:
            return False
        if self.range != value.range:
            return False
        if len(self.children) != len(value.children):
            return False
        for i in range(len(self.children)):
            if self.children[i] != value.children[i]:
                return False
        return True

    @property
    def name(self) -> str:
        return self.variable_name

    @property
    def needsParentheses(self) -> bool:
        return True

    def replaceVariable(self, old: str, new: str):
        if self.variable_name == old:
            self.variable_name = new
        child: DFGNode
        for child in self.children:
            child.replaceVariable(old, new)

    def copy(self):
        node = DFGNode(self.variable_name)
        node.operation = self.operation
        node.range = self.range
        for child in self.children:
            node.children.append(child.copy())
        return node


from enum import Enum


class OPType(Enum):
    BINARY_AND = "binary_and"
    BINARY_OR = "binary_or"
    BINARY_BITAND = "binary_bitand"
    BINARY_XOR = "binary_xor"
    BINARY_XNOR = "binary_xnor"
    BINARY_BITOR = "binary_bitor"
    BINARY_EQ = "binary_eq"
    BINARY_NEQ = "binary_neq"
    BINARY_EQ_EXT = "binary_eq_ext"
    BINARY_NEQ_EXT = "binary_neq_ext"
    BINARY_LT = "binary_lt"
    BINARY_GT = "binary_gt"
    BINARY_LEQ = "binary_leq"
    BINARY_GEQ = "binary_geq"
    BINARY_RSHIFT = "binary_rshift"
    BINARY_RSHIFT_EXT = "binary_rshift_ext"
    BINARY_LSHIFT = "binary_lshift"
    BINARY_LSHIFT_EXT = "binary_lshift_ext"

    # Arithmetic operations
    BINARY_ADD = "binary_add"
    BINARY_SUB = "binary_sub"
    BINARY_MUL = "binary_mul"
    BINARY_DIV = "binary_div"
    BINARY_MOD = "binary_mod"
    BINARY_POW = "binary_pow"

    # Unary operations
    UNARY_POS = "unary_pos"
    UNARY_NEG = "unary_neg"
    UNARY_NOT = "unary_not"
    UNARY_INV = "unary_inv"
    UNARY_AND = "unary_and"
    UNARY_OR = "unary_or"
    UNARY_XOR = "unary_xor"
    UNARY_NAND = "unary_nand"
    UNARY_NOR = "unary_nor"
    UNARY_XNOR = "unary_xnor"
    CONDITIONAL_EXPRESSION = "conditional_expression"

    VARIABLE = "variable"
    CONSTANT = "constant"
    MACRO = "macro"

    ARRAY_CONCAT = "array_concat"
    ARRAY_REPLICATE = "array_replicate"
    ARRAY_SLICE = "array_slice"
    ARRAY_INDEX = "array_index"

    FUNCTION_CALL = "function_call"

    EVENT_INIT = "event_init"
    EVENT_COMB = "event_comb"

    UNARY_POSEDGE = "unary_posedge"
    UNARY_NEGEDGE = "unary_negedge"

    BINARY_EVENT_AND = "binary_event_and"
    BINARY_EVENT_OR = "binary_event_or"

    EVENT_ALWAYS = "event_always"

    UNKNOWN = "unknown"


class OPNode(DFGNode):
    def __init__(self, op: str, operation: OPType, *children) -> None:
        super().__init__(op)
        self.operation = operation
        self.children = children

    def toString(self) -> str:
        opName = self.operation.value
        if opName.startswith("binary_"):
            return f"({self.children[0].toString()} {self.variable_name} {self.children[1].toString()})"
        if opName.startswith("unary_"):
            return f"({self.variable_name} {self.children[0].toString()})"
        match self.operation:
            # Case 1
            # Array operations
            case OPType.ARRAY_CONCAT:
                return f"{{{', '.join([child.toString() for child in self.children])}}}"
            case OPType.ARRAY_REPLICATE:
                return (
                    f"{{{self.children[0].toString()}{{self.children[1].toString()}}}}"
                )
            case OPType.ARRAY_SLICE:
                return f"{self.children[0].toString()}[{self.children[1].toString()}:{self.children[2].toString()}]"
            case OPType.ARRAY_INDEX:
                return f"{self.children[0].toString()}[{self.children[1].toString()}]"
            # Case 2
            # Event operations
            case OPType.EVENT_ALWAYS:
                return f"always @({self.children[0].toString()})"
            case OPType.EVENT_INIT:
                return "initial"
            case OPType.EVENT_COMB:
                return "*"
            # Case 3
            # Special operations
            case OPType.CONDITIONAL_EXPRESSION:
                return f"({self.children[0].toString()} ? {self.children[1].toString()} : {self.children[2].toString()})"
            case OPType.FUNCTION_CALL:
                return f"{self.variable_name}({', '.join([child.toString() for child in self.children])})"
            case OPType.MACRO:
                return f"`{self.children[0].toString()}"
            case OPType.UNKNOWN:
                return self.variable_name
            case _:
                print(f"Unsupported operation {self.operation}")
                raise Exception(f"Unsupported operation {self.operation}")


class VarNode(DFGNode):
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.operation = OPType.VARIABLE

    def toString(self) -> str:
        return (
            self.variable_name
            if self.range is None
            else f"{self.variable_name}{rangeToString(self.range)}"
        )


class ConstantNode(DFGNode):
    def __init__(self, value: Any) -> None:
        super().__init__(str(value))
        self.operation = OPType.CONSTANT

    def toString(self) -> str:
        return self.variable_name


class EmptyEvent(DFGNode):
    def __init__(self) -> None:
        super().__init__("*")
        self.operation = OPType.EVENT_COMB

    def toString(self) -> str:
        return "*"


class InitEvent(DFGNode):
    def __init__(self) -> None:
        super().__init__("initial")
        self.operation = OPType.EVENT_INIT

    def toString(self) -> str:
        return "initial"


class UndefinedNode(DFGNode):
    def __init__(self) -> None:
        super().__init__("*")
        self.operation = OPType.UNKNOWN

    def toString(self) -> str:
        raise Exception("Undefined node cannot be converted to string")



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-28 11:36:22
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-28 12:45:25
"""

from lark import Transformer, v_args
from ...modules import *


class ExpressionTransformer(Transformer):

    def placeholder(self, items):
        return items[0]

    lhs_expression = expression = event_expression = event_parenthesis = placeholder

    def constant_node(self, items):
        return ConstantNode(items[0])

    # constant nodes
    extended_based_number = regular_number = string = based_number = constant_node

    def macro_usage(self, items):
        return OPNode("`", OPType.MACRO, ConstantNode(items[0]))

    def variable(self, items):
        return VarNode(items[0])

    # Generalized operation helper (for all types of operations)
    def op_node(self, symbol, op_type, items):
        return OPNode(symbol, op_type, *items)

    # Unified operations dictionary (binary, unary, array, and conditional)
    operations = {
        # Binary operations
        OPType.BINARY_AND: "&&",
        OPType.BINARY_OR: "||",
        OPType.BINARY_BITAND: "&",
        OPType.BINARY_XOR: "^",
        OPType.BINARY_XNOR: "^~",
        OPType.BINARY_BITOR: "|",
        OPType.BINARY_EQ: "==",
        OPType.BINARY_NEQ: "!=",
        OPType.BINARY_EQ_EXT: "===",
        OPType.BINARY_NEQ_EXT: "!==",
        OPType.BINARY_LT: "<",
        OPType.BINARY_GT: ">",
        OPType.BINARY_LEQ: "<=",
        OPType.BINARY_GEQ: ">=",
        OPType.BINARY_RSHIFT: ">>",
        OPType.BINARY_RSHIFT_EXT: ">>>",
        OPType.BINARY_LSHIFT: "<<",
        OPType.BINARY_LSHIFT_EXT: "<<<",
        OPType.BINARY_ADD: "+",
        OPType.BINARY_SUB: "-",
        OPType.BINARY_MUL: "*",
        OPType.BINARY_DIV: "/",
        OPType.BINARY_MOD: "%",
        OPType.BINARY_POW: "**",
        # Unary operations
        OPType.UNARY_POS: "+",
        OPType.UNARY_NEG: "-",
        OPType.UNARY_NOT: "!",
        OPType.UNARY_INV: "~",
        OPType.UNARY_AND: "&",
        OPType.UNARY_OR: "|",
        OPType.UNARY_XOR: "^",
        OPType.UNARY_NAND: "~&",
        OPType.UNARY_NOR: "~|",
        OPType.UNARY_XNOR: "~^",
        # Array operations
        OPType.ARRAY_CONCAT: "{{{}}}",
        OPType.ARRAY_REPLICATE: "{n{}}",
        OPType.ARRAY_SLICE: "[]",
        OPType.ARRAY_INDEX: "[]",
        # Conditional expression
        OPType.CONDITIONAL_EXPRESSION: "?:",
        # Event expressions
        OPType.UNARY_POSEDGE: "posedge",
        OPType.UNARY_NEGEDGE: "negedge",
        OPType.BINARY_EVENT_OR: "or",
        OPType.BINARY_EVENT_AND: "and",
    }

    # Dynamically generate methods for all operations
    for op_type, symbol in operations.items():
        locals()[f"{op_type.name.lower()}"] = (
            lambda self, items, s=symbol, t=op_type: self.op_node(s, t, items)
        )

    def function_call(self, items):
        return OPNode("()", OPType.FUNCTION_CALL, VarNode(items[0]), *items[1])

    def function_parameters(self, items):
        return items

    def dollar_identifier(self, items):
        return str(items[0])

    def event_variable(self, items):
        return VarNode(items[0])

    def always_block(self, items):
        event_condition: DFGNode = OPNode("always", OPType.EVENT_ALWAYS, items[0])
        assert isinstance(event_condition, DFGNode)
        return [x.setEvent(event_condition) for x in items[1]]

    def initial_block(self, items):
        event_condition: DFGNode = InitEvent()
        assert isinstance(event_condition, DFGNode)
        return [x.setEvent(event_condition) for x in items[0]]

    def combinational_event_expression(self, _):
        return EmptyEvent()



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 21:29:47
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:58:04
"""

from enum import Enum
from .dfg import *


class PortDirection(Enum):
    INPUT = "input"
    OUTPUT = "output"
    INOUT = "inout"

    @staticmethod
    def fromString(direction: str):
        if direction == "input":
            return PortDirection.INPUT
        elif direction == "output":
            return PortDirection.OUTPUT
        elif direction == "inout":
            return PortDirection.INOUT
        else:
            return None

    @staticmethod
    def toString(direction: "PortDirection"):
        return direction.value.lower() if direction is not None else None


class PortType(Enum):
    WIRE = "wire"
    REG = "reg"
    INTEGER = "integer"
    REAL = "real"
    TIME = "time"
    PARAMETER = "parameter"
    LOCALPARAM = "localparam"

    @staticmethod
    def fromString(type: str):
        if type == "wire":
            return PortType.WIRE
        elif type == "reg":
            return PortType.REG
        elif type == "integer":
            return PortType.INTEGER
        elif type == "real":
            return PortType.REAL
        elif type == "time":
            return PortType.TIME
        elif type == "parameter":
            return PortType.PARAMETER
        elif type == "localparam":
            return PortType.LOCALPARAM
        else:
            return None

    @staticmethod
    def toString(type: "PortType"):
        return type.value if type is not None else None


class Port:
    def __init__(
        self,
        variable: DFGNode,
    ):
        self.variable: DFGNode = variable
        self.range: Range = None
        self.direction: PortDirection = None
        self.type: PortType = None

    def setRange(self, range: Range):
        if self.range is None and range is not None:
            self.range = range
            return True
        return False

    def setHeader(self, header: tuple):
        direction, type = header
        return self.setDirection(direction) and self.setType(type)

    def setDirection(self, direction: PortDirection):
        if self.direction is None and direction is not None:
            assert isinstance(direction, PortDirection)
            self.direction = direction
            return True
        return False

    def setType(self, type: str):
        if self.type is None and type is not None:
            assert isinstance(type, PortType)
            self.type = type
            return True
        return False

    def setAll(self, direction: PortDirection, type: str, range: Range):
        self.setDirection(direction)
        self.setType(type)
        self.setRange(range)

    def getType(self):
        return self.type

    def getPortName(self):
        return self.variable.name

    def getRange(self):
        return self.range

    def getHeader(self):
        return (self.direction, self.type)

    def __repr__(self):
        direction_str = (
            "(" + self.direction.value + ")" if self.direction is not None else ""
        )
        return f"Port({self.name}, {direction_str}, {self.type}, {self.range})"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Port):
            return False
        if self.variable != value.variable:
            return False
        if self.direction != value.direction:
            return False
        if self.type != value.type:
            return False
        if self.range != value.range:
            return False
        return True

    @property
    def name(self):
        return self.variable.name


def portToString(port: Port):
    portDirString = ""
    if port.direction is not None:
        portDirString = PortDirection.toString(port.direction) + " "
    typeString = (
        PortType.toString(port.type) + " " if port.type is not None else "wire "
    )

    # range is optional, the space is to be consistent with the space in the middle
    rangeString = f"{rangeToString(port.range)} " if port.range is not None else " "
    return f"{portDirString}{typeString}{rangeString}{port.name}"


class BasicPort(Port):
    def __init__(
        self,
        name: str,
        direction: PortDirection,
        type: str = "wire",
        range: tuple = None,
    ):
        super().__init__(name)
        self.setDirection(direction)
        self.setType(type)
        self.setRange(range)


class OutputPort(BasicPort):
    def __init__(
        self,
        name: str,
        type: str = "wire",
        range: tuple = None,
    ):
        super().__init__(name, PortDirection.OUTPUT, type, range)


class InputPort(BasicPort):
    def __init__(
        self,
        name: str,
        type: str = "wire",
        range: tuple = None,
    ):
        super().__init__(name, PortDirection.INPUT, type, range)


class Frame:
    def __init__(self) -> None:
        self.portDefs = {}
        self.ioPorts = set()

    def addPort(self, port: Port):
        name = port.getPortName()
        if name in self.portDefs:
            self.portDefs[name].setRange(port.getRange())
            self.portDefs[name].setHeader(port.getHeader())
        else:
            self.portDefs[name] = port

        if port.direction is not None:
            self.ioPorts.add(name)

    def addPorts(self, ports: list):
        for port in ports:
            self.addPort(port)

    def getIOs(self) -> set:
        return self.ioPorts

    def getPorts(self) -> dict:
        return self.portDefs

    def getPortNames(self) -> list:
        return list(self.portDefs.keys())

    def getPort(self, portName: str):
        return self.portDefs[portName]

    def getVariableType(self, variableName: str):
        return self.portDefs[variableName].getType()

    def getPortsByType(self, portType: PortDirection):
        return [
            port.name for port in self.portDefs.values() if port.direction == portType
        ]

    @property
    def numInputs(self) -> int:
        return len(self.getPortsByType(PortDirection.INPUT))

    @property
    def numOutputs(self) -> int:
        return len(self.getPortsByType(PortDirection.OUTPUT))

    @property
    def numInouts(self) -> int:
        return len(self.getPortsByType(PortDirection.INOUT))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Frame):
            return False
        # compare ports
        for port in self.getPortNames():
            if port not in other.getPortNames():
                print(f"Port {port} not found in other")
                return False
            if self.getPort(port) != other.getPort(port):
                print(f"Port {port} in self is not equal to port {port} in other")
                return False
        return True



from frontend import *


def test_00_basic():

    verilogStr = """
module top (
    input a,
    input b,
    output f
);
endmodule
"""
    netlist = transformVerilogToNetlist(verilogStr)
    module = netlist.getModule("top")
    assert module is not None
    assert module.numInputs == 2, f"module.numInputs = {module.numInputs} != 2"
    assert module.numOutputs == 1, f"module.numOutputs = {module.numOutputs} != 1"


# Test 01
# Define a module with input, output, and wire ports
def test_01_cornercases():
    verilogStr = """
module top (
    input a,
    input b,
    output f
);
    wire w;
    assign w = a & b;
    assign f = w;
endmodule
"""
    netlist = transformVerilogToNetlist(verilogStr)
    module = netlist.getModule("top")
    assert module is not None
    assert module.numInputs == 2, f"module.numInputs = {module.numInputs} != 2"
    assert module.numOutputs == 1, f"module.numOutputs = {module.numOutputs} != 1"


# Test 02
# Define a module with input, output, inout, and output ports
def test_02_case():
    verilogStr = """
module top (
    input a,
    input b,
    c,
    d,
    output f
);
    output c;
    input d;
endmodule
"""
    netlist = transformVerilogToNetlist(verilogStr)
    module = netlist.getModule("top")
    assert module is not None
    assert module.numInputs == 3, f"module.numInputs = {module.numInputs} != 3"
    assert module.numOutputs == 2, f"module.numOutputs = {module.numOutputs} != 2"


# Test 03
# Define a module with input, output, inout, and output ports
def test_03_case():
    verilogStr = """
module top (
    input wire [3:0] a,
    input b,
    [1:0] c,
    d,
    output f
);
    output c;
    input d;
endmodule
"""
    netlist = transformVerilogToNetlist(verilogStr)
    module = netlist.getModule("top")
    assert module is not None
    assert module.numInputs == 3, f"module.numInputs = {module.numInputs} != 3"
    assert module.numOutputs == 2, f"module.numOutputs = {module.numOutputs} != 2"


# Test 04
# Define a module with input, output, inout, and output ports
def test_04_case():
    verilogStr = """
module top (
    input wire [3:0] a,
    input b,
    [1:0] c,
    d,
    output f
);
    output c;
    input d;
    reg [3:0] e;
    wire [3:0] g;
    
    reg h [3:0];
    reg [3:0] i [3:0];
    
    wire [3:0] j [3:0];
endmodule
"""
    netlist = transformVerilogToNetlist(verilogStr)
    module = netlist.getModule("top")
    assert module is not None
    assert module.numInputs == 3, f"module.numInputs = {module.numInputs} != 3"
    assert module.numOutputs == 2, f"module.numOutputs = {module.numOutputs} != 2"


if __name__ == "__main__":
    test_00_basic()
    test_01_cornercases()
    test_02_case()
    test_03_case()
    test_04_case()



# Reference: https://documentation-rp-test.readthedocs.io/en/latest/tutorfpga04.html#operators-precedence

from frontend import *
import os

# change the working directory to the root directory of the project
os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


def getPrioritized(op1: OPType, op2: OPType) -> str:
    # Generate the Verilog string
    module = Module()
    module.setName("top")
    n1: DFGNode = None
    if op1.numOperands == 2:
        # create a binary operation
        module.addPort(InputPort("n1_a"))
        module.addPort(InputPort("n1_b"))
        a = VarNode("n1_a")
        b = VarNode("n1_b")
        n1 = OPNode("*", op1, a, b)

    elif op1.numOperands == 1:
        # create a unary operation
        module.addPort(InputPort("n1"))
        a = VarNode("n1")
        n1 = OPNode("*", op1, a)

    n2: DFGNode = None
    if op2.numOperands == 2:
        # create a binary operation
        module.addPort(InputPort("na_a"))
        a = VarNode("n2_a")
        n2 = OPNode("*", op2, n1, a)

    elif op2.numOperands == 1:
        # create a unary operation
        a = VarNode("n2")
        n2 = OPNode("*", op2, n1)

    # add the output port
    module.addPort(OutputPort("f"))
    f = VarNode("f")
    module.addAssignment(Assignment(f, n2))

    verilogStr = moduleToString(module)

    # we do a trick and remove the parathesis in the assignment
    for line in verilogStr.split("\n"):
        if "assign" in line:
            newLine = line.replace("(", "").replace(")", "")
            verilogStr = verilogStr.replace(line, newLine)
            break

    netlist = transformVerilogToNetlist(verilogStr)
    module = netlist.getModule("top")

    # we check which operator is on the top
    assign: DFGNode = module.getAssignNode("f")
    return assign.operation


# test 01
# corner cases are stored in the examples/verilogs/operators_precedence.v file
def test_01_cornercases():
    # netlist = readVerilog("examples/verilogs/operators_precedence.v")
    # mut1 = netlist.getModule("test1")
    # mut1.exportDOT()
    # mut1.writeDOT("test1.dot")
    tree = parseVerilogToAST(open("examples/verilogs/operators_precedence.v").read())
    print(tree.pretty())


if __name__ == "__main__":
    test_01_cornercases()
    # test_00_precedence()



"""
Copyright (c) 2024 Hanyu Wang <https://github.com/Nozidoali>

Created Date: Friday, August 16th 2024, 8:01:50 am
Author: Hanyu Wang

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS AS
IS AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.

HISTORY:
Date      	By	Comments
----------	---	----------------------------------------------------------
"""


def changeNameInLines(inputLines: list[str], renameClk: bool = True) -> list[str]:
    data = "\n".join(inputLines)
    newData = changeNameInData(data, renameClk)
    return [x + "\n" for x in newData.split("\n")]


def changeName(inputFile: str, outputFile: str, renameClk: bool = True):
    with open(inputFile, "r") as f:
        data = f.read()
        if not data:
            raise ValueError("Empty file")

    newData = changeNameInData(data, renameClk)
    with open(outputFile, "w") as f:
        f.write(newData)


def changeNameInData(data: str, renameClk: bool = True) -> str:
    import re

    old2new = {}

    # remove comments
    data = re.sub(r"//.*", "", data)
    data = re.sub(r"/\*.*\*/", "", data)

    # remove line breaks
    data = re.sub(r"\n", " ", data)

    # search for wire ...; and reg ...;
    wireDefs = re.findall(r"wire\s+(.*?);", data)
    regDefs = re.findall(r"reg\s+(.*?);", data)
    inputDefs = re.findall(r"input\s+(.*?);", data)
    outputDefs = re.findall(r"output\s+(.*?);", data)

    signals = []
    for definitions in wireDefs + regDefs + inputDefs + outputDefs:
        for x in definitions.split(","):
            x = x.strip()
            if " " in x:
                # this means the definition is like [7:0] signal and has to be split
                x = x.split()[1]
            signals.append(x)

    # search for special characters
    for signal in signals:
        newName = signal
        # check for \\ and replace it with _
        newName = re.sub(r"\\", "", newName)
        # check for | and replace it with __
        newName = re.sub(r"\|", "_", newName)
        if newName != signal:
            old2new[signal] = newName

    # search for the clock signal
    if renameClk:
        old2new["clock"] = "clk"

    # check for the module port list
    modulePortList = re.findall(r"module\s+(.*?)\s*\((.*?)\);", data)
    for modulePort in modulePortList:
        data = getNewDefStr(modulePort[1], data, defStr=None, showRange=False)

    # check the input and output definitions
    for defStr in ["input", "output", "reg", "wire"]:
        defs = re.findall(rf"\s{defStr}\s+(.*?);", data)
        for definition in defs:
            data = getNewDefStr(definition, data, defStr=defStr, showRange=True)

    for old, new in old2new.items():
        try:
            data = data.replace(old, new)
        except Exception as e:
            print(f"Error: {e}")
            print(f"Old: {old}")
            print(f"New: {new}")
            # print(f"Data: {data}")
            exit(1)

    # search for the initial begin block and change it to reset
    # we first match the string between initial begin and end
    m = re.search(r"initial begin(.*?)end", data, re.DOTALL)
    if m:
        content = m.group(1)
        newContents = []
        statements = content.split(";")
        for statement in statements:
            # we wrap it with reset begin and end
            if statement.strip() == "":
                continue
            statement = statement.strip()
            # newContent = f"if (reset) begin\n{statement};\nend "
            newContent = f"{statement}; "
            newContents.append(newContent)
        data = data.replace(content, "\n".join(newContents))

        # finally we replace the initial begin with always @(*)
        data = data.replace("initial begin", "always @(reset) begin ")

    # pretty
    # replace more than one space with one space
    data = re.sub(r"\s+", " ", data)
    data = data.replace(";", ";\n")
    data = data.replace("{", "{\n")
    data = data.replace("}", "}\n")
    data = data.replace("endmodule", "endmodule\n")
    data = data.replace("end ", "end\n")
    data = data.replace("\t", " ")

    return data


def getNewDefStr(
    oldDef: str,
    dataStr: str,
    defStr: str = None,
    showRange: bool = True,
    renameClk: bool = True,
):
    import re

    signalWidth = {}
    ports = []
    for x in oldDef.split(","):
        x = x.strip()

        if " " in x:
            # this means the definition is like [7:0] signal and has to be split
            x = x.split(" ")[1]
        ports.append(x)

    newPortDefs = []
    for port in ports:
        port = port.strip()
        port = re.sub(r"\\", "", port)
        port = re.sub(r"\|", "_", port)

        if renameClk:
            if port == "clock":
                port = "clk"

        if port.endswith("]") and "[" in port:
            portName = port.split("[")[0]
            m = re.search(r"\[(\d+)\]", port)
            if m:
                width = int(m.group(1))
                if portName in signalWidth:
                    signalWidth[portName] = max(signalWidth[portName], width)
                else:
                    signalWidth[portName] = width
            port = portName
        newPortDefs.append(port)

    newPortDefs = list(set(newPortDefs))
    newPortDefsStr = []
    for port in newPortDefs:
        if port in signalWidth and showRange:
            newPortDefsStr.append(f"[{signalWidth[port]}:0] {port}")
        else:
            newPortDefsStr.append(port)

    if defStr is None:
        return dataStr.replace(oldDef, ",\n\t".join(newPortDefsStr))
    else:
        return dataStr.replace(oldDef, f";\n{defStr} ".join(newPortDefsStr))



from ..modules import *
import re


def extractGraph(graph: DotGraph, inputs: list, outputs: list, params: dict = {}):
    excludeFSM = params.get("excludeFSM", True)
    removeREG = params.get("removeREG", True)
    # extract graph from the outputs to the inputs

    # initialize a same type as graph
    newGraph = type(graph)()
    visited = set(inputs)

    fsmGraph = graph.graph.get_subgraph("cluster_fsm")
    if excludeFSM and fsmGraph is not None:
        for node in fsmGraph.nodes():
            visited.add(node.name)

    for output in outputs:
        if output not in visited:
            _extractGraphRec(graph, output, visited, newGraph)

    if removeREG:
        oldToNew = {}

        def findNewVar(var: str):
            # union find
            if var not in oldToNew or oldToNew[var] == var:
                oldToNew[var] = var
                return var
            oldToNew[var] = findNewVar(oldToNew[var])
            return oldToNew[var]

        def mergeVar(varFrom: str, varTo: str):
            # union find
            newVar1 = findNewVar(varFrom)
            newVar2 = findNewVar(varTo)
            if newVar1 != newVar2:
                oldToNew[newVar1] = newVar2

        # remove the register assignments
        assignment: Assignment
        for assignment in newGraph.assignments:
            regVar = assignment.target.name

            # we consider the following two cases:
            # 1. the reg is a pipeling register, then _reg_state<num>
            # 2. the reg is a normal register, then _reg
            if not re.match(r".*(_reg_state\d+|_reg)", regVar):
                continue

            # we get its dependencies
            deps = newGraph.getDependencies(regVar, excludeControls=True)
            assert len(deps) == 1, f"regVar: {regVar}, deps: {deps}"
            dep = deps[0]
            mergeVar(regVar, dep)

        # remove the register assignments
        for old, new in oldToNew.items():
            new = findNewVar(new)
            if old != new:
                # replace the old with the new
                for assignment in newGraph.assignments:
                    if assignment is None:
                        continue
                    assignment.replaceVariable(old, new)
                newGraph.removeVariable(old)

        newGraph.updateAssignments()
    return newGraph


def _extractGraphRec(graph: BNGraph, variable: str, visited: set, newGraph: BNGraph):
    # extract the graph recursively
    if variable in visited:
        return
    if not graph.isDefined(variable):
        return
    visited.add(variable)
    deps = graph.getDependencies(variable)
    # print(f"deps: {deps}")
    for assignment in graph.getAssignmentsOf(variable):
        # add the assignment to the new graph
        newGraph.addAssignment(assignment)
    for dep in deps:
        if dep not in visited:
            _extractGraphRec(graph, dep, visited, newGraph)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 19:20:27
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:51:31
"""

from ...modules import *
from .expressionTransformer import *
from .conditionTransformer import *
from .portTransformer import *
from .moduleTransformer import *


class SystemVerilogTransformer(
    ModuleTransformer,
    ConditionTransformer,
    PortTransformer,
    ExpressionTransformer,
):
    # description: (macro_definition | timescale_definition | module)+
    def start(self, items):
        netlist = Netlist()
        for item in items:
            if isinstance(item, Module):
                netlist.addModule(item)
            elif isinstance(item, Statement):
                item(netlist)
            else:
                raise ValueError(f"Unknown item {item}")
        return netlist

    # timescale_definition: "`timescale" time_unit "/" time_precision
    # time_precision: NUMBER time_identifier
    # time_identifier: "s" | "ms" | "us" | "ns" | "ps"
    def timescale_definition(self, items):
        val1, pre1, val2, pre2 = items
        return TimeScaleDefinition(
            TimeValue(int(val1), str(pre1.data)), TimeValue(int(val2), str(pre2.data))
        )

    # macro_definition: "`define" identifier macro_body
    # macro_body: /.+/
    def macro_definition(self, items):
        name = str(items[0])
        value = str(items[1])
        return MacroDefinition(name, value)



from .bnGraph import *
import pygraphviz as pgv


class DotGraph(BNGraph):
    def __init__(self):
        super().__init__()

    def writeDOT(self, filename: str):
        self.graph.write(filename)

    def exportDOT(self, varName: str = None, params: dict = {}):
        self.graph = pgv.AGraph(strict=False, directed=True)
        self.graph.graph_attr["splines"] = "ortho"
        self.graph.graph_attr["rankdir"] = "TB"  # Top-to-bottom layout

        # parse the params
        skipConstants = params.get("skipConstants", True)
        assignmentsToExport = (
            self.assignments if varName is None else self.var2assigns[varName]
        )

        for i, assignment in enumerate(assignmentsToExport):
            if (
                skipConstants
                and assignment.expression.isConstant()
                and assignment.condition is None
            ):
                continue

            id_lhs = self.exportDOTRec(assignment.target)
            if not assignment.expression.isConstant() or not skipConstants:
                id_rhs = self.exportDOTRec(assignment.expression)

                # add the edge from the rhs to the lhs
                self.graph.add_edge(id_rhs, id_lhs, xlabel=f"val{i}")

            # add the edge from the condition to the lhs
            if assignment.condition is not None:
                id_con = self.exportDOTRec(assignment.condition)
                self.graph.add_edge(
                    id_con, id_lhs, style="dashed", color="red", xlabel=f"cond{i}"
                )

        self.graph.layout(prog="dot")

        # Remove reset node
        nodes_list = list(self.graph.nodes())
        for node in nodes_list:
            if node == "reset" or node == "rst":
                self.graph.remove_node(node)

    def exportDOTRec(self, node: DFGNode):
        if node.isVariable():
            name = node.toString()
            label = name
            shape = "box"
        else:
            name = self.graph.number_of_nodes()
            label = node.name
            shape = "ellipse"

        self.graph.add_node(
            name,
            label=label,
            shape=shape,
        )
        for child in node.children:
            child_idx = self.exportDOTRec(child)
            self.graph.add_edge(child_idx, name)
        return name

    @staticmethod
    def _moveToSubgraph(node: pgv.Node, graph: pgv.AGraph, subgraph: pgv.AGraph):
        subgraph.add_node(node.name, **node.attr)

    @staticmethod
    def _highlightNode(node: pgv.Node, color: str = "CYAN"):
        node.attr["color"] = highlightColors_extended[color]["fontcolor"]
        node.attr["style"] = "filled"
        node.attr["fillcolor"] = highlightColors_extended[color]["fillcolor"]
        # text color
        node.attr["fontcolor"] = highlightColors_extended[color]["fontcolor"]

    def colorFrame(self, variables: list, color: str = "CYAN"):
        self.applyToFrame(variables, self._highlightNode, color)

    def applyToFrame(self, variables: list, func: callable):
        # color the nodes reachable from the current node
        assert self.graph is not None
        for node in self.graph.nodes():
            if node.attr["label"] in variables:
                self.applyToSuccessorsRec(node, set(), func)
                self.applyToPredecessorsRec(node, set(), func)

    def applyToSuccessorsRec(self, node: pgv.Node, visited: set, func: callable):
        func(node)
        visited.add(node)
        for child in self.graph.successors(node):
            if child not in visited:
                self.applyToSuccessorsRec(child, visited, func)

    def applyToPredecessorsRec(self, node: pgv.Node, visited: set, func: callable):
        func(node)
        visited.add(node)
        for child in self.graph.predecessors(node):
            if child not in visited:
                self.applyToPredecessorsRec(child, visited, func)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-28 12:22:41
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-28 13:16:39
"""

from .node import *
from .bnGraph import *
from .dotGraph import *
from .extGraph import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-28 23:18:14
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-28 23:19:18
"""

from setuptools import setup, find_packages

setup(
    name="leap-frontend",
    version="0.1.2",
    author="Hanyu Wang",
    description="Yet another extensible compiler.",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(include=["frontend", "frontend.*"]),
)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 17:11:35
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 22:59:35
"""

from .module import Module
from .definitions import *


class Netlist(Macros, TimeScale):
    """
    A netlist is a collection of modules

    :param modules: a list of modules
    :type modules: list
    :param definitions: a dictionary of definitions
    :type definitions: dict
    """

    def __init__(self):
        Macros.__init__(self)
        TimeScale.__init__(self)
        self.modules = {}

    def addModule(self, module: Module):
        self.modules[module.getName()] = module

    def __repr__(self):
        return "Netlist({})".format(self.modules)

    def getModule(self, name: str) -> Module:
        return self.modules.get(name, None)

    def getModules(self) -> list:
        return list(self.modules.keys())

    def getModuleAt(self, index: int) -> Module:
        # Especially useful for netlist with only one module
        return self.getModule(self.getModules()[index])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Netlist):
            return False
        if len(self.getModules()) != len(other.getModules()):
            print(
                f"Number of modules are not equal: {len(self.getModules())} != {len(other.getModules())}"
            )
            return False
        # compare modules
        for module in self.getModules():
            if module not in other.getModules():
                print(f"Module {module} not found in other")
                print(f"Modules in self: {self.getModules()}")
                print(f"Modules in other: {other.getModules()}")
                return False
            if self.getModule(module) != other.getModule(module):
                print(
                    f"Module {module} in self is not equal to module {module} in other"
                )
                return False

        # compare parameters
        return super().__eq__(other)

    def __ne__(self, value: object) -> bool:
        return not self.__eq__(value)



from frontend import *
import os

# change the working directory to the root directory of the project
os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


def test_00_transformer():
    verilogDir = "examples/verilogs/grammars/"
    for file in os.listdir(verilogDir):
        if file.endswith(".v"):
            netlist: Netlist = readVerilog(verilogDir + file)
            module: Module = netlist.getModule("top")
            assert module is not None
            module.exportDOT()
            module.writeDOT("tmp.dot")
    assert True


if __name__ == "__main__":
    test_00_transformer()



def flatten(items):
    result = []
    for item in items:
        if isinstance(item, list):
            result.extend(flatten(item))
        else:
            if item is not None:
                result.append(item)
    return result



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 18:22:19
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 19:03:09
"""

from .dfg import *


class TimeValue:
    def __init__(self, value: int, unit: str) -> None:
        self.value = value
        self.unit = unit

    def __repr__(self) -> str:
        return f"{int(self.value)} {self.unit}"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, TimeValue):
            print(f"Value {value} is not TimeValue")
            return False
        if self.value != value.value:
            print(f"Value {self.value} != {value.value}")
            return False
        if self.unit != value.unit:
            print(f"Unit {self.unit} != {value.unit}")
            return False
        return True


class Parameter:
    def __init__(self, range: Range, name: str, value: str):
        self.name = name
        self.range = range
        self.value = value

    def __repr__(self):
        return "#{}".format(self.name, self.value)

    def getName(self):
        return self.name


class ParameterHandler:
    def __init__(self):
        self.param_list = {}

    def addParameter(self, param: Parameter):
        self.param_list[param.getName()] = param

    def addParameters(self, parameters: list):
        if not parameters:
            return
        for parameter in parameters:
            self.addParameter(parameter)

    def getParameter(self, name: str):
        return self.param_list[name]

    def getParameters(self):
        return self.param_list

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ParameterHandler):
            return False
        for key, param in self.param_list.items():
            if key not in value.getParameters():
                return False
            if value.getParameter(key) != param:
                return False
        return True


class Macros(dict):
    def __init__(self):
        super().__init__()

    def addMacro(self, name: str, value: str):
        self[name] = value

    def getMacros(self):
        return self

    def getMacroString(self):
        retString = ""
        for key, value in self.items():
            retString += f"`define {key} {value}\n"
        return retString

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Macros):
            return False
        for key, macro in self.items():
            if key not in value.getMacros():
                return False
            if value.getMacros()[key] != macro:
                return False
        return True


class TimeScale:
    def __init__(self):
        self.val1: TimeValue = None
        self.val2: TimeValue = None

    def setTimescale(self, val1: TimeValue, val2: TimeValue):
        self.val1 = val1
        self.val2 = val2

    def getTimescaleString(self):
        if self.val1 is None or self.val2 is None:
            return ""
        return f"`timescale {self.val1} / {self.val2}\n"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, TimeScale):
            return False
        if self.val1 != value.val1:
            return False
        if self.val2 != value.val2:
            return False
        return True



from .node import *

highlightColors_extended = {
    "CYAN": {"fillcolor": "#B3E5FC", "fontcolor": "#0288D1"},  # Light cyan, Dark cyan
    "INDIGO": {
        "fillcolor": "#C5CAE9",
        "fontcolor": "#303F9F",
    },  # Light indigo, Dark indigo
    "VIOLET": {
        "fillcolor": "#D1C4E9",
        "fontcolor": "#512DA8",
    },  # Light violet, Dark violet
    "TEAL": {"fillcolor": "#B2DFDB", "fontcolor": "#00796B"},  # Light teal, Dark teal
    "GREY_BLUE": {
        "fillcolor": "#B0BEC5",
        "fontcolor": "#37474F",
    },  # Light grey-blue, Dark grey-blue
    "GREEN": {
        "fillcolor": "#C8E6C9",
        "fontcolor": "#388E3C",
    },  # Light green, Dark green
    "PURPLE": {
        "fillcolor": "#E1BEE7",
        "fontcolor": "#7B1FA2",
    },  # Light purple, Dark purple
    "ORANGE": {
        "fillcolor": "#FFE0B2",
        "fontcolor": "#FF6F00",
    },  # Light orange, Dark orange
}


class BNGraph:
    def __init__(self):
        self.assignments = []
        self.var2assigns = {}
        self.graph = None

    def addAssignment(self, assignment):
        self.assignments.append(assignment)
        assert isinstance(
            assignment.target, DFGNode
        ), f"Expected DFGNode, got {assignment.target}, {type(assignment.target)}"
        variableName = assignment.target.name
        if variableName not in self.var2assigns:
            self.var2assigns[variableName] = []
        self.var2assigns[variableName].append(len(self.assignments) - 1)

    def removeAssignment(self, idx: int):
        self.assignments[idx] = None

    def removeVariable(self, variableName: str):
        for idx in self.var2assigns[variableName]:
            self.removeAssignment(idx)
        del self.var2assigns[variableName]

    def updateAssignments(self):
        self.assignments = [
            assignment for assignment in self.assignments if assignment is not None
        ]
        self.var2assigns = {}
        for idx, assignment in enumerate(self.assignments):
            variableName = assignment.target.name
            if variableName not in self.var2assigns:
                self.var2assigns[variableName] = []
            self.var2assigns[variableName].append(idx)

    def isDefined(self, variableName: str):
        return variableName in self.var2assigns

    def getAssignmentsOf(self, variableName: str):
        if variableName not in self.var2assigns:
            return []
        return [self.assignments[idx] for idx in self.var2assigns[variableName]]

    def getAssignNode(self, variableName: str):
        # get the node corresponding to the variable
        for assignment in self.getAssignmentsOf(variableName):
            # should be only one
            return assignment.expression
        return None

    def getDependencies(self, variableName: str, excludeControls=False):
        # get the dependencies of the variable
        dependencies = []
        for assignment in self.getAssignmentsOf(variableName):
            root = assignment.expression
            cond = assignment.condition
            dependencies.extend(self.getDependenciesRec(root))
            if not excludeControls and cond is not None:
                dependencies.extend(self.getDependenciesRec(cond))
        return dependencies

    def getDependenciesRec(self, node: DFGNode):
        # get the dependencies of the node
        if node.isVariable():
            return [node.name]
        dependencies = []
        for child in node.children:
            dependencies.extend(self.getDependenciesRec(child))
        return dependencies

    def expandAssignment(self):
        # extract the assignment of the variable, merge the predecessors
        # until we reach the inputs
        for assignment in self.assignments:
            root = assignment.expression
            newNode = self._expandAssignmentRec(root)
            assignment.expression = newNode

    def _expandAssignmentRec(self, node: DFGNode) -> DFGNode:
        # expand the assignment recursively
        if node.isVariable():
            depNode = self.getAssignNode(node.name)
            if depNode is not None:
                return self._expandAssignmentRec(depNode)
            return node.copy()

        newNode: DFGNode = node.copy()
        children = []
        for child in node.children:
            childNode = self._expandAssignmentRec(child)
            children.append(childNode)
        newNode.children = children
        return newNode

    def getVariableNames(self):
        return list(self.var2assigns.keys())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BNGraph):
            return False
        # compare assignments
        for var in self.getVariableNames():
            if var not in other.getVariableNames():
                print(f"Variable {var} not found in other")
                return False
            assignments1 = self.getAssignmentsOf(var)
            assignments2 = other.getAssignmentsOf(var)
            assert len(assignments1) == len(
                assignments2
            ), f"Assignment lengths are not equal: {len(assignments1)} != {len(assignments2)}"
            for i in range(len(assignments1)):
                if assignments1[i] != assignments2[i]:
                    print(
                        f"Assignment {i} is not equal, {assignments1[i]} != {assignments2[i]}"
                    )
                    return False
        return True

    def __ne__(self, value: object) -> bool:
        return not self.__eq__(value)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-19 00:28:31
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-23 23:20:41
"""

from ..modules import *
from .headerWriter import *
from .moduleWriter import *


def writeVerilogAST(module: Module, filename: str):
    assert isinstance(module, Module), "module should be an instance of Module"
    module.dfg.toGraph(filename)


def netlistToString(netlist: Netlist) -> str:
    verilogStr = ""
    verilogStr += netlist.getMacroString() + netlist.getTimescaleString() + "\n"

    for moduleName in netlist.getModules():
        module = netlist.getModule(moduleName)
        verilogStr += moduleToString(module) + "\n"

    return verilogStr


def writeVerilog(netlist: Netlist | Module, filename: str):
    """
    Write the netlist to a verilog file

    :param netlist: the netlist or module to write
    :type netlist: Netlist | Module
    :param filename: the name of the file to write
    :type filename: str

    .. note::
        The input can be either a netlist or a module

    Example usage:

    .. code-block:: python

        netlist = readVerilog("path/to/verilog/file.v")
        writeVerilog(netlist, "path/to/output/file.v")

    .. seealso::

        Related functions: :func:`readVerilog`
    """
    assert isinstance(netlist, Netlist) or isinstance(
        netlist, Module
    ), "netlist should be an instance of Netlist or Module"
    with open(filename, "w") as f:
        if isinstance(netlist, Module):
            f.write(moduleToString(netlist))
        else:
            f.write(netlistToString(netlist))



from frontend import *
import pytest


def write_after_read(data: str):
    netlist: Netlist = transformVerilogToNetlist(data)
    newData = netlistToString(netlist)
    print(newData)
    newNetlist: Netlist = transformVerilogToNetlist(newData)
    return netlist == newNetlist


# Test 00
# Basic Module Definition
def test_00_write_assignment():
    verilogString = """
module top(
        a,
        f,
        b
);
input  a;
input  b;
output  f;
endmodule
    """
    assert write_after_read(verilogString)


# Test 01
# Module Definition with bad port names
def test_01_write_assignment():
    verilogString = """
module top (
    input wire [3:0] a,
    input b,
    [1:0] c,
    d,
    output f
);
    output c;
    input [1:0] d;
endmodule
"""
    assert write_after_read(verilogString)


# Test 02
# Module Definition with bad port names
@pytest.mark.skip(reason="not implemented")
def test_02_write_assignment():
    verilogString = """
module top (
    input wire [3:0] a,
    input b,
    [1:0] c,
    d,
    output f
);
    output c;
    input d;
    reg [3:0] e;
    wire [3:0] g;
    
    reg h [3:0];
    reg [3:0] i [3:0];
    
    wire [3:0] j [3:0];
endmodule
"""
    assert write_after_read(verilogString)


# Test 03
# Basic assignment
def test_03_write_assignment():
    verilogString = """
module top (
    input wire [3:0] a,
    output f
);
    assign f = &a;
endmodule
"""
    assert write_after_read(verilogString)


# Test 04
# Basic assignment with expression
def test_04_write_assignment():
    verilogString = """
module top (
    input wire [3:0] a,
    output f
);
    assign f = &a ^ 1;
endmodule
"""
    assert write_after_read(verilogString)


# Test 05
# Basic assignment with Macro
def test_05_write_assignment():
    verilogString = """
module top (
    input wire [3:0] a,
    output f
);
    assign f = &a ^ `ONE;
endmodule
"""
    assert write_after_read(verilogString)


# Test 06
# Range with Macro
def test_06_write_assignment():
    verilogString = """
module top (
    input wire [`FOUR-1:0] a,
    output f
);
    assign f = &a ^ `ONE;
endmodule
"""
    assert write_after_read(verilogString)


# Test 07
# Expression Test Case 1
def test_07_write_expression():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output f
);
    assign f = (&a) ^ b;
endmodule
"""
    assert write_after_read(verilogString)


# Test 08
# Binary Operator Test - AND, OR, XOR
def test_08_binary_operations():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output wire f
);
    assign f = (a && b) || (a ^ b);
endmodule
"""
    assert write_after_read(verilogString)


# Test 09
# Unary Operator Test - NOT, INV
def test_09_unary_operations():
    verilogString = """
module top (
    input wire [3:0] a,
    output wire f
);
    assign f = !a | ~a;
endmodule
"""
    assert write_after_read(verilogString)


# Test 10
# Shift Operators - Left and Right Shift
def test_10_shift_operations():
    verilogString = """
module top (
    input wire [3:0] a,
    output wire f
);
    assign f = a << 2;
endmodule
"""
    assert write_after_read(verilogString)


# Test 11
# Array Concatenation
def test_11_array_concat():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output wire [7:0] f
);
    assign f = {a, b};
endmodule
"""
    assert write_after_read(verilogString)


# Test 12
# Conditional Expression
def test_12_conditional_expression():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output wire f
);
    assign f = (a > b) ? 1 : 0;
endmodule
"""
    assert write_after_read(verilogString)


# Test 13
# Multiple Binary Operations in a Single Expression
def test_13_multiple_operations():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output wire f
);
    assign f = (a + b) - (a * b);
endmodule
"""
    assert write_after_read(verilogString)


# Test 14
# With if conditions
def test_14_conditions():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output wire f
);
    if (a > b) begin
        assign f = 1;
    end else begin
        assign f = 0;
    end

endmodule
"""
    assert write_after_read(verilogString)


# Test 15
# Blocking and Non-Blocking Assignments
def test_15_conditions():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output wire f,
    output reg g
);
    if (a > b) begin
        assign f = 1;
    end else begin
        assign f = 0;
    end
    
    if (a > b) begin
        g <= a;
    end else begin
        g <= b;
    end

endmodule
"""
    assert write_after_read(verilogString)


# Test 16
# Always event
def test_16_conditions():
    verilogString = """
module top (
    input wire clk,
    input wire [3:0] a,
    input wire [3:0] b,
    output wire f,
    output reg g
);

always @(posedge clk) begin
    if (a > b) begin
        assign f = 1;
    end else begin
        assign f = 0;
    end
    
    if (a > b) begin
        g <= a;
    end else begin
        g <= b;
    end
end
endmodule
"""
    assert write_after_read(verilogString)


# Test 17
# Always event
def test_17_conditions():
    verilogString = """
module top (
    input clk,
    input wire [3:0] a,
    input wire [3:0] b,
    output wire f,
    output reg g
);

always @(negedge clk) begin
    if (a > b) begin
        assign f = 1;
    end else begin
        assign f = 0;
    end
    
    if (a > b) begin
        g <= a;
    end else begin
        g <= b;
    end
end
endmodule
"""
    assert write_after_read(verilogString)


# Test 18
# Reset and positive edge clock event
def test_18_reset_event():
    verilogString = """
module top (
    input clk,
    input reset,
    input wire [3:0] a,
    input wire [3:0] b,
    output reg f
);

always @(posedge clk or posedge reset) begin
    if (reset) begin
        f <= 0;
    end else begin
        f <= a + b;
    end
end
endmodule
"""
    assert write_after_read(verilogString)


# Test 19
# Initial block with event-driven assignment
def test_19_initial_event():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output reg f
);

initial begin
    f = 0;
end

always @(a or b) begin
    f <= a + b;
end
endmodule
"""
    assert write_after_read(verilogString)


# Test 20
# Always block with multiple conditions and mixed operators (and, or)
def test_20_mixed_event_conditions():
    verilogString = """
module top (
    input clk,
    input reset,
    input wire [3:0] a,
    input wire [3:0] b,
    output reg f
);

always @(posedge clk or negedge reset or a or b) begin
    if (!reset) begin
        f <= 0;
    end else if (a && b) begin
        f <= 1;
    end else if (a || b) begin
        f <= a + b;
    end else begin
        f <= 0;
    end
end
endmodule
"""
    assert write_after_read(verilogString)


# Test 21
# Mixed event-driven assignments with priority for * (wildcard)
def test_21_star_event():
    verilogString = """
module top (
    input wire [3:0] a,
    input wire [3:0] b,
    output reg f
);

always @(*) begin
    if (a == b) begin
        f <= 1;
    end else begin
        f <= 0;
    end
end
endmodule
"""
    assert write_after_read(verilogString)


# Test 22
# Multiple events combining and and or
@pytest.mark.skip(reason="not implemented")
def test_22_and_or_event():
    verilogString = """
module top (
    input clk,
    input wire [3:0] a,
    input wire [3:0] b,
    output reg f
);

always @(posedge clk or (a && b)) begin
    f <= (a && b) ? 1 : 0;
end
endmodule
"""
    assert write_after_read(verilogString)


# Test 23
# Parameters
def test_23_parameters():
    verilogString = """
`define WIDTH 4
module top (
    input wire [`WIDTH -1:0] a,
    input wire [`WIDTH -1:0] b,
    output wire [`WIDTH -1:0] f
);

assign f = a + b;
endmodule
"""
    assert write_after_read(verilogString)


# Test 24
# Timescale
def test_24_parameters():
    verilogString = """
`define WIDTH 4
`timescale 1ns / 1ps
module top (
    input wire [`WIDTH -1:0] a,
    input wire [`WIDTH -1:0] b,
    output wire [`WIDTH -1:0] f
);

assign f = a + b;
endmodule
"""
    assert write_after_read(verilogString)


# Test 25
# Parameters inside modules
def test_25_parameters():
    verilogString = """
module top (
    input wire a,
    input wire b,
    output wire f
);

`define WIDTH 4

wire [`WIDTH -1:0] a;
wire [`WIDTH -1:0] b;
wire [`WIDTH -1:0] f;

assign f = a + b;
endmodule
"""
    assert write_after_read(verilogString)


if __name__ == "__main__":
    # test_00_write_assignment()
    # test_01_write_assignment()
    # test_02_write_assignment()
    # test_03_write_assignment()
    # test_04_write_assignment()
    # test_05_write_assignment()
    # test_06_write_assignment()
    # test_07_write_expression()
    # test_08_binary_operations()
    # test_09_unary_operations()
    # test_10_shift_operations()
    # test_11_array_concat()
    # test_12_conditional_expression()
    # test_13_multiple_operations()
    # test_14_conditions()
    # test_15_conditions()
    # test_16_conditions()
    # test_17_conditions()
    # test_18_reset_event()
    # test_19_initial_event()
    # test_20_mixed_event_conditions()
    # test_21_star_event()
    # test_22_and_or_event()
    # test_23_parameters()
    # test_24_parameters()
    test_25_parameters()
